from __future__ import annotations

import asyncio
import inspect
import logging
import time  # Added for measuring duration
from typing import Any, Callable, Iterable, Optional, Set, Tuple, Union, overload

from typing_extensions import TypeAlias

from chalk.features.tag import Environments
from chalk.utils.collections import ensure_tuple
from chalk.utils.log_with_context import get_logger

HookFn: TypeAlias = Callable[[], Any]


_hook_logger = get_logger("chalk.hook_logger")


async def _run_all_hooks(environment: str, hooks: Iterable["Hook"]) -> None:
    for hook in hooks:
        if hook.environment is None or environment in hook.environment:
            start_time = time.perf_counter()  # Start timing
            try:
                _hook_logger.info("Starting to run hook %s...", hook.fn.__name__)
                if inspect.iscoroutinefunction(hook.fn):
                    await hook()
                else:
                    await asyncio.get_running_loop().run_in_executor(None, hook)
            except Exception as e:
                logging.error(f"Error running hook {hook.fn.__name__}", exc_info=True)
                raise e
            finally:
                duration = time.perf_counter() - start_time  # Calculate duration
                _hook_logger.info("Ran hook %s in %.2f seconds", hook.fn.__name__, duration)


class Hook:
    # Registry
    before_all: Set["Hook"] = set()
    after_all: Set["Hook"] = set()

    environment: Optional[Tuple[str, ...]]
    fn: HookFn
    filename: str

    def __init__(self, fn: HookFn, filename: str, environment: Optional[Environments] = None):
        super().__init__()
        self.fn = fn
        self.filename = filename
        self.environment = None if environment is None else ensure_tuple(environment)

    def __call__(self):
        return self.fn()

    def __repr__(self):
        return f'Hook(filename={self.filename}, fn={self.fn.__name__}", environment={str(self.environment)})'

    @classmethod
    async def async_run_all_before_all(cls, environment: str) -> None:
        return await _run_all_hooks(environment, cls.before_all)

    @classmethod
    async def async_run_all_after_all(cls, environment: str) -> None:
        return await _run_all_hooks(environment, cls.after_all)


@overload
def before_all(fn: HookFn, /) -> Hook:
    ...


@overload
def before_all(fn: None = None, /, environment: Optional[Environments] = None) -> Callable[[HookFn], Hook]:
    ...


def before_all(
    fn: Optional[HookFn] = None, /, environment: Optional[Environments] = None
) -> Union[Hook, Callable[[HookFn], Hook]]:
    def decorator(f: HookFn):
        caller_filename = inspect.getsourcefile(f) or "unknown_file"
        hook = Hook(fn=f, filename=caller_filename, environment=environment)
        Hook.before_all.add(hook)
        return hook

    return decorator(fn) if fn else decorator


@overload
def after_all(fn: HookFn, /, environment: Optional[Environments] = None) -> Hook:
    ...


@overload
def after_all(fn: None = None, /, environment: Optional[Environments] = None) -> Callable[[HookFn], Hook]:
    ...


def after_all(
    fn: Optional[HookFn] = None, /, environment: Optional[Environments] = None
) -> Union[Hook, Callable[[HookFn], Hook]]:
    def decorator(f: HookFn):
        caller_filename = inspect.getsourcefile(f) or "unknown_file"
        hook = Hook(fn=f, filename=caller_filename, environment=environment)
        Hook.after_all.add(hook)
        return hook

    return decorator(fn) if fn else decorator



from chalk._gen.chalk.aggregate.v1 import timeseries_pb2 as _timeseries_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class AggregateBackfillCostEstimate(_message.Message):
    __slots__ = ("max_buckets", "expected_buckets", "expected_bytes", "expected_storage_cost", "expected_runtime")
    MAX_BUCKETS_FIELD_NUMBER: _ClassVar[int]
    EXPECTED_BUCKETS_FIELD_NUMBER: _ClassVar[int]
    EXPECTED_BYTES_FIELD_NUMBER: _ClassVar[int]
    EXPECTED_STORAGE_COST_FIELD_NUMBER: _ClassVar[int]
    EXPECTED_RUNTIME_FIELD_NUMBER: _ClassVar[int]
    max_buckets: int
    expected_buckets: int
    expected_bytes: int
    expected_storage_cost: float
    expected_runtime: _duration_pb2.Duration
    def __init__(
        self,
        max_buckets: _Optional[int] = ...,
        expected_buckets: _Optional[int] = ...,
        expected_bytes: _Optional[int] = ...,
        expected_storage_cost: _Optional[float] = ...,
        expected_runtime: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
    ) -> None: ...

class AggregateBackfillUserParams(_message.Message):
    __slots__ = ("features", "resolver", "timestamp_column_name", "lower_bound", "upper_bound", "exact")
    FEATURES_FIELD_NUMBER: _ClassVar[int]
    RESOLVER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_COLUMN_NAME_FIELD_NUMBER: _ClassVar[int]
    LOWER_BOUND_FIELD_NUMBER: _ClassVar[int]
    UPPER_BOUND_FIELD_NUMBER: _ClassVar[int]
    EXACT_FIELD_NUMBER: _ClassVar[int]
    features: _containers.RepeatedScalarFieldContainer[str]
    resolver: str
    timestamp_column_name: str
    lower_bound: _timestamp_pb2.Timestamp
    upper_bound: _timestamp_pb2.Timestamp
    exact: bool
    def __init__(
        self,
        features: _Optional[_Iterable[str]] = ...,
        resolver: _Optional[str] = ...,
        timestamp_column_name: _Optional[str] = ...,
        lower_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        upper_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        exact: bool = ...,
    ) -> None: ...

class AggregateBackfill(_message.Message):
    __slots__ = (
        "series",
        "resolver",
        "datetime_feature",
        "bucket_duration",
        "filters_description",
        "group_by",
        "max_retention",
        "lower_bound",
        "upper_bound",
    )
    SERIES_FIELD_NUMBER: _ClassVar[int]
    RESOLVER_FIELD_NUMBER: _ClassVar[int]
    DATETIME_FEATURE_FIELD_NUMBER: _ClassVar[int]
    BUCKET_DURATION_FIELD_NUMBER: _ClassVar[int]
    FILTERS_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    GROUP_BY_FIELD_NUMBER: _ClassVar[int]
    MAX_RETENTION_FIELD_NUMBER: _ClassVar[int]
    LOWER_BOUND_FIELD_NUMBER: _ClassVar[int]
    UPPER_BOUND_FIELD_NUMBER: _ClassVar[int]
    series: _containers.RepeatedCompositeFieldContainer[_timeseries_pb2.AggregateTimeSeries]
    resolver: str
    datetime_feature: str
    bucket_duration: _duration_pb2.Duration
    filters_description: str
    group_by: _containers.RepeatedScalarFieldContainer[str]
    max_retention: _duration_pb2.Duration
    lower_bound: _timestamp_pb2.Timestamp
    upper_bound: _timestamp_pb2.Timestamp
    def __init__(
        self,
        series: _Optional[_Iterable[_Union[_timeseries_pb2.AggregateTimeSeries, _Mapping]]] = ...,
        resolver: _Optional[str] = ...,
        datetime_feature: _Optional[str] = ...,
        bucket_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        filters_description: _Optional[str] = ...,
        group_by: _Optional[_Iterable[str]] = ...,
        max_retention: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        lower_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        upper_bound: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class AggregateBackfillWithCostEstimate(_message.Message):
    __slots__ = ("backfill", "estimate")
    BACKFILL_FIELD_NUMBER: _ClassVar[int]
    ESTIMATE_FIELD_NUMBER: _ClassVar[int]
    backfill: AggregateBackfill
    estimate: AggregateBackfillCostEstimate
    def __init__(
        self,
        backfill: _Optional[_Union[AggregateBackfill, _Mapping]] = ...,
        estimate: _Optional[_Union[AggregateBackfillCostEstimate, _Mapping]] = ...,
    ) -> None: ...

class AggregateBackfillJob(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "resolver",
        "features",
        "agent_id",
        "deployment_id",
        "created_at",
        "updated_at",
        "resolvers",
        "cron_aggregate_backfill_id",
        "plan_hash",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    RESOLVER_FIELD_NUMBER: _ClassVar[int]
    FEATURES_FIELD_NUMBER: _ClassVar[int]
    AGENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    RESOLVERS_FIELD_NUMBER: _ClassVar[int]
    CRON_AGGREGATE_BACKFILL_ID_FIELD_NUMBER: _ClassVar[int]
    PLAN_HASH_FIELD_NUMBER: _ClassVar[int]
    id: str
    environment_id: str
    resolver: str
    features: _containers.RepeatedScalarFieldContainer[str]
    agent_id: str
    deployment_id: str
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    resolvers: _containers.RepeatedScalarFieldContainer[str]
    cron_aggregate_backfill_id: str
    plan_hash: str
    def __init__(
        self,
        id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        resolver: _Optional[str] = ...,
        features: _Optional[_Iterable[str]] = ...,
        agent_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        resolvers: _Optional[_Iterable[str]] = ...,
        cron_aggregate_backfill_id: _Optional[str] = ...,
        plan_hash: _Optional[str] = ...,
    ) -> None: ...

class CronAggregateBackfill(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "deployment_id",
        "schedule",
        "plan_hash",
        "features",
        "resolvers",
        "created_at",
        "updated_at",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    SCHEDULE_FIELD_NUMBER: _ClassVar[int]
    PLAN_HASH_FIELD_NUMBER: _ClassVar[int]
    FEATURES_FIELD_NUMBER: _ClassVar[int]
    RESOLVERS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    id: str
    environment_id: str
    deployment_id: str
    schedule: str
    plan_hash: str
    features: _containers.RepeatedScalarFieldContainer[str]
    resolvers: _containers.RepeatedScalarFieldContainer[str]
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    def __init__(
        self,
        id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        schedule: _Optional[str] = ...,
        plan_hash: _Optional[str] = ...,
        features: _Optional[_Iterable[str]] = ...,
        resolvers: _Optional[_Iterable[str]] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...



from chalk._gen.chalk.arrow.v1 import arrow_pb2 as _arrow_pb2
from chalk._gen.chalk.common.v1 import operation_kind_pb2 as _operation_kind_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureValuesTimestampType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FEATURE_VALUES_TIMESTAMP_TYPE_UNSPECIFIED: _ClassVar[FeatureValuesTimestampType]
    FEATURE_VALUES_TIMESTAMP_TYPE_INSERTED_AT: _ClassVar[FeatureValuesTimestampType]
    FEATURE_VALUES_TIMESTAMP_TYPE_OBSERVED_AT: _ClassVar[FeatureValuesTimestampType]

FEATURE_VALUES_TIMESTAMP_TYPE_UNSPECIFIED: FeatureValuesTimestampType
FEATURE_VALUES_TIMESTAMP_TYPE_INSERTED_AT: FeatureValuesTimestampType
FEATURE_VALUES_TIMESTAMP_TYPE_OBSERVED_AT: FeatureValuesTimestampType

class FeatureValueFilters(_message.Message):
    __slots__ = ("resolver_fqn", "deployment_id", "operation_id", "operation_kind", "primary_key")
    RESOLVER_FQN_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    OPERATION_KIND_FIELD_NUMBER: _ClassVar[int]
    PRIMARY_KEY_FIELD_NUMBER: _ClassVar[int]
    resolver_fqn: _containers.RepeatedScalarFieldContainer[str]
    deployment_id: _containers.RepeatedScalarFieldContainer[str]
    operation_id: _containers.RepeatedScalarFieldContainer[str]
    operation_kind: _containers.RepeatedScalarFieldContainer[_operation_kind_pb2.OperationKind]
    primary_key: _containers.RepeatedCompositeFieldContainer[_arrow_pb2.ScalarValue]
    def __init__(
        self,
        resolver_fqn: _Optional[_Iterable[str]] = ...,
        deployment_id: _Optional[_Iterable[str]] = ...,
        operation_id: _Optional[_Iterable[str]] = ...,
        operation_kind: _Optional[_Iterable[_Union[_operation_kind_pb2.OperationKind, str]]] = ...,
        primary_key: _Optional[_Iterable[_Union[_arrow_pb2.ScalarValue, _Mapping]]] = ...,
    ) -> None: ...

class GetFeatureValuesPageToken(_message.Message):
    __slots__ = ("timestamp_hwm", "operation_id_hwm", "observation_id_hwm")
    TIMESTAMP_HWM_FIELD_NUMBER: _ClassVar[int]
    OPERATION_ID_HWM_FIELD_NUMBER: _ClassVar[int]
    OBSERVATION_ID_HWM_FIELD_NUMBER: _ClassVar[int]
    timestamp_hwm: _timestamp_pb2.Timestamp
    operation_id_hwm: str
    observation_id_hwm: str
    def __init__(
        self,
        timestamp_hwm: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        operation_id_hwm: _Optional[str] = ...,
        observation_id_hwm: _Optional[str] = ...,
    ) -> None: ...

class GetFeatureValuesRequest(_message.Message):
    __slots__ = (
        "feature_fqn",
        "timestamp_type",
        "lower_bound_inclusive",
        "upper_bound_exclusive",
        "filters",
        "page_size",
        "page_token",
    )
    FEATURE_FQN_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_TYPE_FIELD_NUMBER: _ClassVar[int]
    LOWER_BOUND_INCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    UPPER_BOUND_EXCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    PAGE_SIZE_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    feature_fqn: str
    timestamp_type: FeatureValuesTimestampType
    lower_bound_inclusive: _timestamp_pb2.Timestamp
    upper_bound_exclusive: _timestamp_pb2.Timestamp
    filters: FeatureValueFilters
    page_size: int
    page_token: str
    def __init__(
        self,
        feature_fqn: _Optional[str] = ...,
        timestamp_type: _Optional[_Union[FeatureValuesTimestampType, str]] = ...,
        lower_bound_inclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        upper_bound_exclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        filters: _Optional[_Union[FeatureValueFilters, _Mapping]] = ...,
        page_size: _Optional[int] = ...,
        page_token: _Optional[str] = ...,
    ) -> None: ...

class GetFeatureValuesResponse(_message.Message):
    __slots__ = ("next_page_token", "total_size", "parquet")
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    TOTAL_SIZE_FIELD_NUMBER: _ClassVar[int]
    PARQUET_FIELD_NUMBER: _ClassVar[int]
    next_page_token: str
    total_size: int
    parquet: bytes
    def __init__(
        self, next_page_token: _Optional[str] = ..., total_size: _Optional[int] = ..., parquet: _Optional[bytes] = ...
    ) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import prompt_pb2 as chalk_dot_server_dot_v1_dot_prompt__pb2


class PromptServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListNamedPrompts = channel.unary_unary(
            "/chalk.server.v1.PromptService/ListNamedPrompts",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListNamedPromptsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListNamedPromptsResponse.FromString,
        )
        self.GetNamedPrompt = channel.unary_unary(
            "/chalk.server.v1.PromptService/GetNamedPrompt",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetNamedPromptRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetNamedPromptResponse.FromString,
        )
        self.CreateNamedPrompt = channel.unary_unary(
            "/chalk.server.v1.PromptService/CreateNamedPrompt",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreateNamedPromptRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreateNamedPromptResponse.FromString,
        )
        self.UpdateNamedPrompt = channel.unary_unary(
            "/chalk.server.v1.PromptService/UpdateNamedPrompt",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.UpdateNamedPromptRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.UpdateNamedPromptResponse.FromString,
        )
        self.ListPromptVariants = channel.unary_unary(
            "/chalk.server.v1.PromptService/ListPromptVariants",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptVariantsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptVariantsResponse.FromString,
        )
        self.ListPromptEvaluationRuns = channel.unary_unary(
            "/chalk.server.v1.PromptService/ListPromptEvaluationRuns",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptEvaluationRunsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptEvaluationRunsResponse.FromString,
        )
        self.GetPromptEvaluationRun = channel.unary_unary(
            "/chalk.server.v1.PromptService/GetPromptEvaluationRun",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetPromptEvaluationRunRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetPromptEvaluationRunResponse.FromString,
        )
        self.CreatePromptEvaluationRun = channel.unary_unary(
            "/chalk.server.v1.PromptService/CreatePromptEvaluationRun",
            request_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreatePromptEvaluationRunRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreatePromptEvaluationRunResponse.FromString,
        )


class PromptServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ListNamedPrompts(self, request, context):
        """Return a list of named prompts"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetNamedPrompt(self, request, context):
        """Return a single named prompt with the latest prompt variant"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateNamedPrompt(self, request, context):
        """Create a new named prompt and its corresponding prompt variant"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def UpdateNamedPrompt(self, request, context):
        """Modify a named prompt; may create a new prompt variant if variant settings changed"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListPromptVariants(self, request, context):
        """Return a list of prompt variants"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListPromptEvaluationRuns(self, request, context):
        """Return a list of prompt evaluation runs"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetPromptEvaluationRun(self, request, context):
        """Return a prompt evaluation run and its prompt variants"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreatePromptEvaluationRun(self, request, context):
        """Start a new prompt evaluation run"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_PromptServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "ListNamedPrompts": grpc.unary_unary_rpc_method_handler(
            servicer.ListNamedPrompts,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListNamedPromptsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListNamedPromptsResponse.SerializeToString,
        ),
        "GetNamedPrompt": grpc.unary_unary_rpc_method_handler(
            servicer.GetNamedPrompt,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetNamedPromptRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetNamedPromptResponse.SerializeToString,
        ),
        "CreateNamedPrompt": grpc.unary_unary_rpc_method_handler(
            servicer.CreateNamedPrompt,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreateNamedPromptRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreateNamedPromptResponse.SerializeToString,
        ),
        "UpdateNamedPrompt": grpc.unary_unary_rpc_method_handler(
            servicer.UpdateNamedPrompt,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.UpdateNamedPromptRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.UpdateNamedPromptResponse.SerializeToString,
        ),
        "ListPromptVariants": grpc.unary_unary_rpc_method_handler(
            servicer.ListPromptVariants,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptVariantsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptVariantsResponse.SerializeToString,
        ),
        "ListPromptEvaluationRuns": grpc.unary_unary_rpc_method_handler(
            servicer.ListPromptEvaluationRuns,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptEvaluationRunsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptEvaluationRunsResponse.SerializeToString,
        ),
        "GetPromptEvaluationRun": grpc.unary_unary_rpc_method_handler(
            servicer.GetPromptEvaluationRun,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetPromptEvaluationRunRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.GetPromptEvaluationRunResponse.SerializeToString,
        ),
        "CreatePromptEvaluationRun": grpc.unary_unary_rpc_method_handler(
            servicer.CreatePromptEvaluationRun,
            request_deserializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreatePromptEvaluationRunRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_prompt__pb2.CreatePromptEvaluationRunResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.PromptService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class PromptService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ListNamedPrompts(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/ListNamedPrompts",
            chalk_dot_server_dot_v1_dot_prompt__pb2.ListNamedPromptsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.ListNamedPromptsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetNamedPrompt(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/GetNamedPrompt",
            chalk_dot_server_dot_v1_dot_prompt__pb2.GetNamedPromptRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.GetNamedPromptResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateNamedPrompt(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/CreateNamedPrompt",
            chalk_dot_server_dot_v1_dot_prompt__pb2.CreateNamedPromptRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.CreateNamedPromptResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def UpdateNamedPrompt(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/UpdateNamedPrompt",
            chalk_dot_server_dot_v1_dot_prompt__pb2.UpdateNamedPromptRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.UpdateNamedPromptResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListPromptVariants(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/ListPromptVariants",
            chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptVariantsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptVariantsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListPromptEvaluationRuns(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/ListPromptEvaluationRuns",
            chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptEvaluationRunsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.ListPromptEvaluationRunsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetPromptEvaluationRun(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/GetPromptEvaluationRun",
            chalk_dot_server_dot_v1_dot_prompt__pb2.GetPromptEvaluationRunRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.GetPromptEvaluationRunResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreatePromptEvaluationRun(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.PromptService/CreatePromptEvaluationRun",
            chalk_dot_server_dot_v1_dot_prompt__pb2.CreatePromptEvaluationRunRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_prompt__pb2.CreatePromptEvaluationRunResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from __future__ import annotations

import ast
import importlib.machinery
import inspect
import linecache
import os
import sys
from typing import Any, Callable, Optional, Type


class _ChalkClassFoundException(Exception):
    def __init__(self, node: ast.ClassDef):
        super().__init__()
        self.node = node


class _ChalkFunctionFoundException(Exception):
    def __init__(self, node: ast.FunctionDef | ast.AsyncFunctionDef):
        super().__init__()
        self.node = node


class _ChalkClassFinder(ast.NodeVisitor):
    def __init__(self, qualname: str):
        super().__init__()
        self.stack = []
        self.qualname = qualname

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.stack.append(node.name)
        self.stack.append("<locals>")
        self.generic_visit(node)
        self.stack.pop()
        self.stack.pop()

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.stack.append(node.name)
        self.stack.append("<locals>")
        self.generic_visit(node)
        self.stack.pop()
        self.stack.pop()

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.stack.append(node.name)
        if self.qualname == ".".join(self.stack):
            # Return the decorator for the class if present
            if node.decorator_list:
                line_number = node.decorator_list[0].lineno
            else:
                line_number = node.lineno

            # decrement by one since lines starts with indexing by zero
            line_number -= 1
            raise _ChalkClassFoundException(node=node)
        self.generic_visit(node)
        self.stack.pop()


class _ChalkFunctionFinder(ast.NodeVisitor):
    def __init__(self, name: str):
        super().__init__()
        self.stack = []
        self.name = name

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.stack.append(node.name)
        if self.name == ".".join(self.stack):
            raise _ChalkFunctionFoundException(node=node)
        self.generic_visit(node)
        self.stack.pop()

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.stack.append(node.name)
        if self.name == ".".join(self.stack):
            raise _ChalkFunctionFoundException(node=node)
        self.generic_visit(node)
        self.stack.pop()

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.stack.append(node.name)
        self.generic_visit(node)
        self.stack.pop()


def get_class_ast(cls: Type) -> ast.ClassDef:
    if not inspect.isclass(cls):
        raise ValueError(f"Expected class, but found {cls}")

    file = get_source_filename(cls)
    if file:
        # Invalidate cache if needed.
        linecache.checkcache(file)
    else:
        file = inspect.getfile(cls)
        # Allow filenames in form of "<something>" to pass through.
        # `doctest` monkeypatches `linecache` module to enable
        # inspection, so let `linecache.getlines` to be called.
        if not (file.startswith("<") and file.endswith(">")):
            raise OSError("source code not available")

    module = inspect.getmodule(cls, file)
    if module:
        lines = linecache.getlines(file, module.__dict__)
    else:
        lines = linecache.getlines(file)
    if not lines:
        raise OSError("could not get source code")

    qualname = cls.__qualname__
    source = "".join(lines)
    tree = ast.parse(source)
    class_finder = _ChalkClassFinder(qualname)
    try:
        class_finder.visit(tree)
    except _ChalkClassFoundException as e:
        return e.node
    else:
        raise OSError("could not find class definition")


def get_function_ast(fn: Callable) -> ast.FunctionDef | ast.AsyncFunctionDef:
    file = get_source_filename(fn)
    if file:
        # Invalidate cache if needed.
        linecache.checkcache(file)
    else:
        file = inspect.getfile(fn)
        # Allow filenames in form of "<something>" to pass through.
        # `doctest` monkeypatches `linecache` module to enable
        # inspection, so let `linecache.getlines` to be called.
        if not (file.startswith("<") and file.endswith(">")):
            raise OSError("source code not available")

    module = inspect.getmodule(fn, file)
    if module:
        lines = linecache.getlines(file, module.__dict__)
    else:
        lines = linecache.getlines(file)
    if not lines:
        raise OSError("could not get source code")

    name = fn.__name__  # not a class so I think we should use the name
    source = "".join(lines)
    tree = ast.parse(source)
    function_finder = _ChalkFunctionFinder(name)
    try:
        function_finder.visit(tree)
    except _ChalkFunctionFoundException as e:
        return e.node
    else:
        raise OSError("could not find function definition")


def get_source_filename(obj: Any) -> Optional[str]:
    """Modified in first line from inspect.getsource(obj)"""
    try:
        filename = sys.modules[obj.__module__].__file__
    except:
        filename = inspect.getfile(obj)
    if filename is None:
        return None
    all_bytecode_suffixes = importlib.machinery.DEBUG_BYTECODE_SUFFIXES[:]
    all_bytecode_suffixes += importlib.machinery.OPTIMIZED_BYTECODE_SUFFIXES[:]
    if any(filename.endswith(s) for s in all_bytecode_suffixes):
        filename = os.path.splitext(filename)[0] + importlib.machinery.SOURCE_SUFFIXES[0]
    elif any(filename.endswith(s) for s in importlib.machinery.EXTENSION_SUFFIXES):
        return None
    if os.path.exists(filename):
        return filename
    # only return a non-existent filename if the module has a PEP 302 loader
    module = inspect.getmodule(obj, filename)
    if getattr(module, "__loader__", None) is not None:
        return filename
    elif getattr(getattr(module, "__spec__", None), "loader", None) is not None:
        return filename
    # or it is in the linecache
    elif filename in linecache.cache:
        return filename
    return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/utils/v1/sensitive.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n\x1e\x63halk/utils/v1/sensitive.proto\x12\x0e\x63halk.utils.v1\x1a google/protobuf/descriptor.proto:@\n\tsensitive\x12\x1d.google.protobuf.FieldOptions\x18\x9b\xf4\x04 \x01(\x08R\tsensitive\x88\x01\x01\x42~\n\x12\x63om.chalk.utils.v1B\x0eSensitiveProtoP\x01\xa2\x02\x03\x43UX\xaa\x02\x0e\x43halk.Utils.V1\xca\x02\x0e\x43halk\\Utils\\V1\xe2\x02\x1a\x43halk\\Utils\\V1\\GPBMetadata\xea\x02\x10\x43halk::Utils::V1b\x06proto3"
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.utils.v1.sensitive_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\022com.chalk.utils.v1B\016SensitiveProtoP\001\242\002\003CUX\252\002\016Chalk.Utils.V1\312\002\016Chalk\\Utils\\V1\342\002\032Chalk\\Utils\\V1\\GPBMetadata\352\002\020Chalk::Utils::V1"
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/lsp/v1/lsp.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x16\x63halk/lsp/v1/lsp.proto\x12\x0c\x63halk.lsp.v1"\x83\x01\n\x03LSP\x12H\n\x0b\x64iagnostics\x18\x01 \x03(\x0b\x32&.chalk.lsp.v1.PublishDiagnosticsParamsR\x0b\x64iagnostics\x12\x32\n\x07\x61\x63tions\x18\x02 \x03(\x0b\x32\x18.chalk.lsp.v1.CodeActionR\x07\x61\x63tions"h\n\x18PublishDiagnosticsParams\x12\x10\n\x03uri\x18\x01 \x01(\tR\x03uri\x12:\n\x0b\x64iagnostics\x18\x02 \x03(\x0b\x32\x18.chalk.lsp.v1.DiagnosticR\x0b\x64iagnostics"\xd8\x02\n\nDiagnostic\x12)\n\x05range\x18\x01 \x01(\x0b\x32\x13.chalk.lsp.v1.RangeR\x05range\x12\x18\n\x07message\x18\x02 \x01(\tR\x07message\x12<\n\x08severity\x18\x03 \x01(\x0e\x32 .chalk.lsp.v1.DiagnosticSeverityR\x08severity\x12\x17\n\x04\x63ode\x18\x04 \x01(\tH\x00R\x04\x63ode\x88\x01\x01\x12H\n\x10\x63ode_description\x18\x05 \x01(\x0b\x32\x1d.chalk.lsp.v1.CodeDescriptionR\x0f\x63odeDescription\x12[\n\x13related_information\x18\x06 \x03(\x0b\x32*.chalk.lsp.v1.DiagnosticRelatedInformationR\x12relatedInformationB\x07\n\x05_code"_\n\x05Range\x12,\n\x05start\x18\x01 \x01(\x0b\x32\x16.chalk.lsp.v1.PositionR\x05start\x12(\n\x03\x65nd\x18\x02 \x01(\x0b\x32\x16.chalk.lsp.v1.PositionR\x03\x65nd"]\n\x08Position\x12\x17\n\x04line\x18\x01 \x01(\x05H\x00R\x04line\x88\x01\x01\x12!\n\tcharacter\x18\x02 \x01(\x05H\x01R\tcharacter\x88\x01\x01\x42\x07\n\x05_lineB\x0c\n\n_character"\x8f\x01\n\nCodeAction\x12\x14\n\x05title\x18\x01 \x01(\tR\x05title\x12:\n\x0b\x64iagnostics\x18\x02 \x03(\x0b\x32\x18.chalk.lsp.v1.DiagnosticR\x0b\x64iagnostics\x12/\n\x04\x65\x64it\x18\x03 \x01(\x0b\x32\x1b.chalk.lsp.v1.WorkspaceEditR\x04\x65\x64it"Z\n\rWorkspaceEdit\x12I\n\x10\x64ocument_changes\x18\x01 \x03(\x0b\x32\x1e.chalk.lsp.v1.TextDocumentEditR\x0f\x64ocumentChanges"\x8b\x01\n\x10TextDocumentEdit\x12I\n\rtext_document\x18\x01 \x01(\x0b\x32$.chalk.lsp.v1.TextDocumentIdentifierR\x0ctextDocument\x12,\n\x05\x65\x64its\x18\x02 \x03(\x0b\x32\x16.chalk.lsp.v1.TextEditR\x05\x65\x64its"*\n\x16TextDocumentIdentifier\x12\x10\n\x03uri\x18\x01 \x01(\tR\x03uri"P\n\x08TextEdit\x12)\n\x05range\x18\x01 \x01(\x0b\x32\x13.chalk.lsp.v1.RangeR\x05range\x12\x19\n\x08new_text\x18\x02 \x01(\tR\x07newText"%\n\x0f\x43odeDescription\x12\x12\n\x04href\x18\x01 \x01(\tR\x04href"l\n\x1c\x44iagnosticRelatedInformation\x12\x32\n\x08location\x18\x01 \x01(\x0b\x32\x16.chalk.lsp.v1.LocationR\x08location\x12\x18\n\x07message\x18\x02 \x01(\tR\x07message"G\n\x08Location\x12\x10\n\x03uri\x18\x01 \x01(\tR\x03uri\x12)\n\x05range\x18\x02 \x01(\x0b\x32\x13.chalk.lsp.v1.RangeR\x05range*\xbc\x01\n\x12\x44iagnosticSeverity\x12#\n\x1f\x44IAGNOSTIC_SEVERITY_UNSPECIFIED\x10\x00\x12\x1d\n\x19\x44IAGNOSTIC_SEVERITY_ERROR\x10\x01\x12\x1f\n\x1b\x44IAGNOSTIC_SEVERITY_WARNING\x10\x02\x12#\n\x1f\x44IAGNOSTIC_SEVERITY_INFORMATION\x10\x03\x12\x1c\n\x18\x44IAGNOSTIC_SEVERITY_HINT\x10\x04\x42n\n\x10\x63om.chalk.lsp.v1B\x08LspProtoP\x01\xa2\x02\x03\x43LX\xaa\x02\x0c\x43halk.Lsp.V1\xca\x02\x0c\x43halk\\Lsp\\V1\xe2\x02\x18\x43halk\\Lsp\\V1\\GPBMetadata\xea\x02\x0e\x43halk::Lsp::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.lsp.v1.lsp_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\020com.chalk.lsp.v1B\010LspProtoP\001\242\002\003CLX\252\002\014Chalk.Lsp.V1\312\002\014Chalk\\Lsp\\V1\342\002\030Chalk\\Lsp\\V1\\GPBMetadata\352\002\016Chalk::Lsp::V1"
    _globals["_DIAGNOSTICSEVERITY"]._serialized_start = 1548
    _globals["_DIAGNOSTICSEVERITY"]._serialized_end = 1736
    _globals["_LSP"]._serialized_start = 41
    _globals["_LSP"]._serialized_end = 172
    _globals["_PUBLISHDIAGNOSTICSPARAMS"]._serialized_start = 174
    _globals["_PUBLISHDIAGNOSTICSPARAMS"]._serialized_end = 278
    _globals["_DIAGNOSTIC"]._serialized_start = 281
    _globals["_DIAGNOSTIC"]._serialized_end = 625
    _globals["_RANGE"]._serialized_start = 627
    _globals["_RANGE"]._serialized_end = 722
    _globals["_POSITION"]._serialized_start = 724
    _globals["_POSITION"]._serialized_end = 817
    _globals["_CODEACTION"]._serialized_start = 820
    _globals["_CODEACTION"]._serialized_end = 963
    _globals["_WORKSPACEEDIT"]._serialized_start = 965
    _globals["_WORKSPACEEDIT"]._serialized_end = 1055
    _globals["_TEXTDOCUMENTEDIT"]._serialized_start = 1058
    _globals["_TEXTDOCUMENTEDIT"]._serialized_end = 1197
    _globals["_TEXTDOCUMENTIDENTIFIER"]._serialized_start = 1199
    _globals["_TEXTDOCUMENTIDENTIFIER"]._serialized_end = 1241
    _globals["_TEXTEDIT"]._serialized_start = 1243
    _globals["_TEXTEDIT"]._serialized_end = 1323
    _globals["_CODEDESCRIPTION"]._serialized_start = 1325
    _globals["_CODEDESCRIPTION"]._serialized_end = 1362
    _globals["_DIAGNOSTICRELATEDINFORMATION"]._serialized_start = 1364
    _globals["_DIAGNOSTICRELATEDINFORMATION"]._serialized_end = 1472
    _globals["_LOCATION"]._serialized_start = 1474
    _globals["_LOCATION"]._serialized_end = 1545
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import deploy_pb2 as chalk_dot_server_dot_v1_dot_deploy__pb2


class DeployServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.DeployBranch = channel.unary_unary(
            "/chalk.server.v1.DeployService/DeployBranch",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.DeployBranchRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.DeployBranchResponse.FromString,
        )
        self.CreateBranchFromSourceDeployment = channel.unary_unary(
            "/chalk.server.v1.DeployService/CreateBranchFromSourceDeployment",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.CreateBranchFromSourceDeploymentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.CreateBranchFromSourceDeploymentResponse.FromString,
        )
        self.GetDeployment = channel.unary_unary(
            "/chalk.server.v1.DeployService/GetDeployment",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetDeploymentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetDeploymentResponse.FromString,
        )
        self.ListDeployments = channel.unary_unary(
            "/chalk.server.v1.DeployService/ListDeployments",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ListDeploymentsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ListDeploymentsResponse.FromString,
        )
        self.GetActiveDeployments = channel.unary_unary(
            "/chalk.server.v1.DeployService/GetActiveDeployments",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetActiveDeploymentsRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetActiveDeploymentsResponse.FromString,
        )
        self.SuspendDeployment = channel.unary_unary(
            "/chalk.server.v1.DeployService/SuspendDeployment",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.SuspendDeploymentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.SuspendDeploymentResponse.FromString,
        )
        self.ScaleDeployment = channel.unary_unary(
            "/chalk.server.v1.DeployService/ScaleDeployment",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ScaleDeploymentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ScaleDeploymentResponse.FromString,
        )
        self.TagDeployment = channel.unary_unary(
            "/chalk.server.v1.DeployService/TagDeployment",
            request_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.TagDeploymentRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.TagDeploymentResponse.FromString,
        )


class DeployServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def DeployBranch(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CreateBranchFromSourceDeployment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetDeployment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListDeployments(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetActiveDeployments(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SuspendDeployment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ScaleDeployment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def TagDeployment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_DeployServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "DeployBranch": grpc.unary_unary_rpc_method_handler(
            servicer.DeployBranch,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.DeployBranchRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.DeployBranchResponse.SerializeToString,
        ),
        "CreateBranchFromSourceDeployment": grpc.unary_unary_rpc_method_handler(
            servicer.CreateBranchFromSourceDeployment,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.CreateBranchFromSourceDeploymentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.CreateBranchFromSourceDeploymentResponse.SerializeToString,
        ),
        "GetDeployment": grpc.unary_unary_rpc_method_handler(
            servicer.GetDeployment,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetDeploymentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetDeploymentResponse.SerializeToString,
        ),
        "ListDeployments": grpc.unary_unary_rpc_method_handler(
            servicer.ListDeployments,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ListDeploymentsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ListDeploymentsResponse.SerializeToString,
        ),
        "GetActiveDeployments": grpc.unary_unary_rpc_method_handler(
            servicer.GetActiveDeployments,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetActiveDeploymentsRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.GetActiveDeploymentsResponse.SerializeToString,
        ),
        "SuspendDeployment": grpc.unary_unary_rpc_method_handler(
            servicer.SuspendDeployment,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.SuspendDeploymentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.SuspendDeploymentResponse.SerializeToString,
        ),
        "ScaleDeployment": grpc.unary_unary_rpc_method_handler(
            servicer.ScaleDeployment,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ScaleDeploymentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.ScaleDeploymentResponse.SerializeToString,
        ),
        "TagDeployment": grpc.unary_unary_rpc_method_handler(
            servicer.TagDeployment,
            request_deserializer=chalk_dot_server_dot_v1_dot_deploy__pb2.TagDeploymentRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_deploy__pb2.TagDeploymentResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.DeployService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class DeployService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def DeployBranch(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/DeployBranch",
            chalk_dot_server_dot_v1_dot_deploy__pb2.DeployBranchRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.DeployBranchResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CreateBranchFromSourceDeployment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/CreateBranchFromSourceDeployment",
            chalk_dot_server_dot_v1_dot_deploy__pb2.CreateBranchFromSourceDeploymentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.CreateBranchFromSourceDeploymentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetDeployment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/GetDeployment",
            chalk_dot_server_dot_v1_dot_deploy__pb2.GetDeploymentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.GetDeploymentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListDeployments(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/ListDeployments",
            chalk_dot_server_dot_v1_dot_deploy__pb2.ListDeploymentsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.ListDeploymentsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetActiveDeployments(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/GetActiveDeployments",
            chalk_dot_server_dot_v1_dot_deploy__pb2.GetActiveDeploymentsRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.GetActiveDeploymentsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SuspendDeployment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/SuspendDeployment",
            chalk_dot_server_dot_v1_dot_deploy__pb2.SuspendDeploymentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.SuspendDeploymentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ScaleDeployment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/ScaleDeployment",
            chalk_dot_server_dot_v1_dot_deploy__pb2.ScaleDeploymentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.ScaleDeploymentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def TagDeployment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.DeployService/TagDeployment",
            chalk_dot_server_dot_v1_dot_deploy__pb2.TagDeploymentRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_deploy__pb2.TagDeploymentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from __future__ import annotations

from typing import Any, Mapping, Sequence, Union

import pyarrow as pa

__all__ = ["TJSON", "pyarrow_json_type", "is_pyarrow_json_type", "JSON"]

TJSON = Union[None, str, int, float, bool, Sequence["TJSON"], Mapping[str, "TJSON"]]
JSON = Union[None, str, int, float, bool, Sequence, Mapping[str, object]]


class _JSONType(pa.ExtensionType):  # pyright: ignore[reportUntypedBaseClass]
    # Implementation of https://arrow.apache.org/docs/format/CanonicalExtensions.html#json
    def __init__(self):
        super().__init__(pa.large_utf8(), "arrow.json")

    def __eq__(self, other: Any):
        return isinstance(other, _JSONType)

    def __hash__(self):
        return hash(repr(self))

    def __arrow_ext_serialize__(self):
        # since we don't have a parameterized type, we don't need extra
        # metadata to be deserialized
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: pa.DataType, serialized: bytes):
        # return an instance of this subclass given the serialized
        # metadata.
        return _JSON_TYPE


_JSON_TYPE = _JSONType()

try:
    # If a json type is already registered, unregister it -- we'll use our own
    pa.unregister_extension_type("arrow.json")
except pa.ArrowKeyError:
    pass

pa.register_extension_type(_JSON_TYPE)


def pyarrow_json_type() -> pa.ExtensionType:
    return _JSON_TYPE


def is_pyarrow_json_type(x: pa.DataType):
    return getattr(x, "extension_name", None) == "arrow.json"



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.engine.v2 import (
    feature_values_chart_pb2 as chalk_dot_engine_dot_v2_dot_feature__values__chart__pb2,
)
from chalk._gen.chalk.engine.v2 import feature_values_pb2 as chalk_dot_engine_dot_v2_dot_feature__values__pb2
from chalk._gen.chalk.engine.v2 import query_log_pb2 as chalk_dot_engine_dot_v2_dot_query__log__pb2
from chalk._gen.chalk.engine.v2 import query_values_pb2 as chalk_dot_engine_dot_v2_dot_query__values__pb2


class OfflineStoreServiceStub(object):
    """This service exposes endpoints for dealing with the offline store. It should never depend on the python graph.
    v2 introduces two breaking changes:
    Uses messages from engine.v2 instead of common.v1 (common is not meant for engine-specific messages)
    Removes certain endpoints added and had to be immediately deprecated due to deprecation
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetQueryLogEntries = channel.unary_unary(
            "/chalk.engine.v2.OfflineStoreService/GetQueryLogEntries",
            request_serializer=chalk_dot_engine_dot_v2_dot_query__log__pb2.GetQueryLogEntriesRequest.SerializeToString,
            response_deserializer=chalk_dot_engine_dot_v2_dot_query__log__pb2.GetQueryLogEntriesResponse.FromString,
        )
        self.GetQueryValues = channel.unary_unary(
            "/chalk.engine.v2.OfflineStoreService/GetQueryValues",
            request_serializer=chalk_dot_engine_dot_v2_dot_query__values__pb2.GetQueryValuesRequest.SerializeToString,
            response_deserializer=chalk_dot_engine_dot_v2_dot_query__values__pb2.GetQueryValuesResponse.FromString,
        )
        self.GetFeatureValuesTimeSeriesChart = channel.unary_unary(
            "/chalk.engine.v2.OfflineStoreService/GetFeatureValuesTimeSeriesChart",
            request_serializer=chalk_dot_engine_dot_v2_dot_feature__values__chart__pb2.GetFeatureValuesTimeSeriesChartRequest.SerializeToString,
            response_deserializer=chalk_dot_engine_dot_v2_dot_feature__values__chart__pb2.GetFeatureValuesTimeSeriesChartResponse.FromString,
        )
        self.GetFeatureValues = channel.unary_unary(
            "/chalk.engine.v2.OfflineStoreService/GetFeatureValues",
            request_serializer=chalk_dot_engine_dot_v2_dot_feature__values__pb2.GetFeatureValuesRequest.SerializeToString,
            response_deserializer=chalk_dot_engine_dot_v2_dot_feature__values__pb2.GetFeatureValuesResponse.FromString,
        )


class OfflineStoreServiceServicer(object):
    """This service exposes endpoints for dealing with the offline store. It should never depend on the python graph.
    v2 introduces two breaking changes:
    Uses messages from engine.v2 instead of common.v1 (common is not meant for engine-specific messages)
    Removes certain endpoints added and had to be immediately deprecated due to deprecation
    """

    def GetQueryLogEntries(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetQueryValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetFeatureValuesTimeSeriesChart(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetFeatureValues(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_OfflineStoreServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetQueryLogEntries": grpc.unary_unary_rpc_method_handler(
            servicer.GetQueryLogEntries,
            request_deserializer=chalk_dot_engine_dot_v2_dot_query__log__pb2.GetQueryLogEntriesRequest.FromString,
            response_serializer=chalk_dot_engine_dot_v2_dot_query__log__pb2.GetQueryLogEntriesResponse.SerializeToString,
        ),
        "GetQueryValues": grpc.unary_unary_rpc_method_handler(
            servicer.GetQueryValues,
            request_deserializer=chalk_dot_engine_dot_v2_dot_query__values__pb2.GetQueryValuesRequest.FromString,
            response_serializer=chalk_dot_engine_dot_v2_dot_query__values__pb2.GetQueryValuesResponse.SerializeToString,
        ),
        "GetFeatureValuesTimeSeriesChart": grpc.unary_unary_rpc_method_handler(
            servicer.GetFeatureValuesTimeSeriesChart,
            request_deserializer=chalk_dot_engine_dot_v2_dot_feature__values__chart__pb2.GetFeatureValuesTimeSeriesChartRequest.FromString,
            response_serializer=chalk_dot_engine_dot_v2_dot_feature__values__chart__pb2.GetFeatureValuesTimeSeriesChartResponse.SerializeToString,
        ),
        "GetFeatureValues": grpc.unary_unary_rpc_method_handler(
            servicer.GetFeatureValues,
            request_deserializer=chalk_dot_engine_dot_v2_dot_feature__values__pb2.GetFeatureValuesRequest.FromString,
            response_serializer=chalk_dot_engine_dot_v2_dot_feature__values__pb2.GetFeatureValuesResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.engine.v2.OfflineStoreService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class OfflineStoreService(object):
    """This service exposes endpoints for dealing with the offline store. It should never depend on the python graph.
    v2 introduces two breaking changes:
    Uses messages from engine.v2 instead of common.v1 (common is not meant for engine-specific messages)
    Removes certain endpoints added and had to be immediately deprecated due to deprecation
    """

    @staticmethod
    def GetQueryLogEntries(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.engine.v2.OfflineStoreService/GetQueryLogEntries",
            chalk_dot_engine_dot_v2_dot_query__log__pb2.GetQueryLogEntriesRequest.SerializeToString,
            chalk_dot_engine_dot_v2_dot_query__log__pb2.GetQueryLogEntriesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetQueryValues(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.engine.v2.OfflineStoreService/GetQueryValues",
            chalk_dot_engine_dot_v2_dot_query__values__pb2.GetQueryValuesRequest.SerializeToString,
            chalk_dot_engine_dot_v2_dot_query__values__pb2.GetQueryValuesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetFeatureValuesTimeSeriesChart(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.engine.v2.OfflineStoreService/GetFeatureValuesTimeSeriesChart",
            chalk_dot_engine_dot_v2_dot_feature__values__chart__pb2.GetFeatureValuesTimeSeriesChartRequest.SerializeToString,
            chalk_dot_engine_dot_v2_dot_feature__values__chart__pb2.GetFeatureValuesTimeSeriesChartResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetFeatureValues(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.engine.v2.OfflineStoreService/GetFeatureValues",
            chalk_dot_engine_dot_v2_dot_feature__values__pb2.GetFeatureValuesRequest.SerializeToString,
            chalk_dot_engine_dot_v2_dot_feature__values__pb2.GetFeatureValuesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



from functools import cache
from typing import Any

import pyarrow as pa
from google.protobuf.descriptor import Descriptor, FieldDescriptor, FileDescriptor
from google.protobuf.descriptor_pb2 import FileDescriptorProto, FileDescriptorSet
from google.protobuf.descriptor_pool import DescriptorPool
from google.protobuf.message import Message

from chalk._gen.chalk.arrow.v1 import arrow_pb2 as pb

PROTOBUF_TO_UNIT = {
    pb.TIME_UNIT_SECOND: "s",
    pb.TIME_UNIT_MILLISECOND: "ms",
    pb.TIME_UNIT_MICROSECOND: "us",
    pb.TIME_UNIT_NANOSECOND: "ns",
}


UNIT_TO_PROTOBUF = {
    "s": pb.TIME_UNIT_SECOND,
    "ms": pb.TIME_UNIT_MILLISECOND,
    "us": pb.TIME_UNIT_MICROSECOND,
    "ns": pb.TIME_UNIT_NANOSECOND,
}


def _get_pyarrow_type_for_proto_field(field_descriptor: FieldDescriptor) -> pa.DataType:
    """Maps protobuf primitive field types to PyArrow types."""
    type_mapping = {
        FieldDescriptor.TYPE_DOUBLE: pa.float64(),
        FieldDescriptor.TYPE_FLOAT: pa.float32(),
        FieldDescriptor.TYPE_INT64: pa.int64(),
        FieldDescriptor.TYPE_UINT64: pa.uint64(),
        FieldDescriptor.TYPE_INT32: pa.int32(),
        FieldDescriptor.TYPE_FIXED64: pa.uint64(),
        FieldDescriptor.TYPE_FIXED32: pa.uint32(),
        FieldDescriptor.TYPE_BOOL: pa.bool_(),
        FieldDescriptor.TYPE_STRING: pa.large_utf8(),
        FieldDescriptor.TYPE_BYTES: pa.binary(),
        FieldDescriptor.TYPE_UINT32: pa.uint32(),
        FieldDescriptor.TYPE_SFIXED32: pa.int32(),
        FieldDescriptor.TYPE_SFIXED64: pa.int64(),
        FieldDescriptor.TYPE_SINT32: pa.int32(),
        FieldDescriptor.TYPE_SINT64: pa.int64(),
        FieldDescriptor.TYPE_ENUM: pa.int32(),  # Could be refined further
    }
    return type_mapping.get(field_descriptor.type, pa.null())


@cache
def convert_proto_message_type_to_pyarrow_type(
    proto_message_class: Descriptor, proto_max_depth: int = 15
) -> pa.DataType:
    """Converts a Protocol Buffer message class into an equivalent PyArrow struct type."""

    def _convert_proto_message_type_to_pyarrow_type(
        proto_message_class: Descriptor, max_depth: int, self_referential_detector: set[Descriptor]
    ) -> pa.DataType:
        """Converts a Protocol Buffer message class into an equivalent PyArrow struct type."""
        if max_depth == 0:
            raise RecursionError(
                f"Recursion limit exceeded when converting proto message type to pyarrow. This error occurs if the resulting pyarrow structure would be over {proto_max_depth} levels deep."
            )
        struct_fields = []
        for fd in proto_message_class.fields:
            if fd.type == FieldDescriptor.TYPE_MESSAGE:
                new_message_type: Descriptor = fd.message_type
                if new_message_type in self_referential_detector:
                    raise RecursionError(
                        f"Infinitely recursive proto structure detected when converting proto message type to pyarrow - message '{new_message_type.full_name}' has a self-referential definition."
                    )
                self_referential_detector.add(new_message_type)
                field_type = _convert_proto_message_type_to_pyarrow_type(
                    new_message_type, max_depth=max_depth - 1, self_referential_detector=self_referential_detector
                )
                self_referential_detector.remove(new_message_type)
            else:
                field_type = _get_pyarrow_type_for_proto_field(fd)
            if fd.label == FieldDescriptor.LABEL_REPEATED:
                field_type = pa.large_list(field_type)
            struct_fields.append(pa.field(fd.name, field_type, fd.label != FieldDescriptor.LABEL_REQUIRED))
        return pa.struct(struct_fields)

    return _convert_proto_message_type_to_pyarrow_type(
        proto_message_class, max_depth=proto_max_depth, self_referential_detector=set()
    )


def create_empty_pyarrow_scalar_from_proto_type(proto_message: Message) -> pa.Scalar:
    """Creates a PyArrow scalar with None value but with type structure matching the Protobuf message."""
    pa_type = convert_proto_message_type_to_pyarrow_type(proto_message.DESCRIPTOR)
    return pa.scalar({}, type=pa_type)


@cache
def reconstruct_message_field_descriptor(serialized: bytes, full_name: str):
    file_descriptor_set = FileDescriptorSet()
    file_descriptor_set.ParseFromString(serialized)

    descriptor_pool = DescriptorPool()
    for file in file_descriptor_set.file:
        descriptor_pool.Add(file)

    return descriptor_pool.FindMessageTypeByName(full_name)


@cache
def serialize_message_file_descriptor(file_descriptor: Any) -> bytes:
    """Create a FileDescriptorSet containing the given Protobuf message file descriptor and its dependencies."""
    file_descriptor_set = FileDescriptorSet()
    processed_files = set()

    def add_file_and_deps(fd: FileDescriptor):
        if fd.name in processed_files:
            return

        for dep in fd.dependencies:
            add_file_and_deps(dep)

        fd_proto = FileDescriptorProto()
        fd.CopyToProto(fd_proto)

        file_descriptor_set.file.append(fd_proto)
        processed_files.add(fd.name)

    add_file_and_deps(file_descriptor)
    return file_descriptor_set.SerializeToString()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

from chalk._gen.chalk.server.v1 import authtesting_pb2 as chalk_dot_server_dot_v1_dot_authtesting__pb2


class AuthTestingServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetUnauthedTestEndpoint = channel.unary_unary(
            "/chalk.server.v1.AuthTestingService/GetUnauthedTestEndpoint",
            request_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetUnauthedTestEndpointRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetUnauthedTestEndpointResponse.FromString,
        )
        self.GetAuthedTestEndpoint = channel.unary_unary(
            "/chalk.server.v1.AuthTestingService/GetAuthedTestEndpoint",
            request_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAuthedTestEndpointRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAuthedTestEndpointResponse.FromString,
        )
        self.GetViewerTestEndpoint = channel.unary_unary(
            "/chalk.server.v1.AuthTestingService/GetViewerTestEndpoint",
            request_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetViewerTestEndpointRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetViewerTestEndpointResponse.FromString,
        )
        self.GetDataScientistTestEndpoint = channel.unary_unary(
            "/chalk.server.v1.AuthTestingService/GetDataScientistTestEndpoint",
            request_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDataScientistTestEndpointRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDataScientistTestEndpointResponse.FromString,
        )
        self.GetDeveloperTestEndpoint = channel.unary_unary(
            "/chalk.server.v1.AuthTestingService/GetDeveloperTestEndpoint",
            request_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDeveloperTestEndpointRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDeveloperTestEndpointResponse.FromString,
        )
        self.GetAdminTestEndpoint = channel.unary_unary(
            "/chalk.server.v1.AuthTestingService/GetAdminTestEndpoint",
            request_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAdminTestEndpointRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAdminTestEndpointResponse.FromString,
        )
        self.GetOwnerTestEndpoint = channel.unary_unary(
            "/chalk.server.v1.AuthTestingService/GetOwnerTestEndpoint",
            request_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetOwnerTestEndpointRequest.SerializeToString,
            response_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetOwnerTestEndpointResponse.FromString,
        )


class AuthTestingServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetUnauthedTestEndpoint(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAuthedTestEndpoint(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetViewerTestEndpoint(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetDataScientistTestEndpoint(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetDeveloperTestEndpoint(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetAdminTestEndpoint(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetOwnerTestEndpoint(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_AuthTestingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetUnauthedTestEndpoint": grpc.unary_unary_rpc_method_handler(
            servicer.GetUnauthedTestEndpoint,
            request_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetUnauthedTestEndpointRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetUnauthedTestEndpointResponse.SerializeToString,
        ),
        "GetAuthedTestEndpoint": grpc.unary_unary_rpc_method_handler(
            servicer.GetAuthedTestEndpoint,
            request_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAuthedTestEndpointRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAuthedTestEndpointResponse.SerializeToString,
        ),
        "GetViewerTestEndpoint": grpc.unary_unary_rpc_method_handler(
            servicer.GetViewerTestEndpoint,
            request_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetViewerTestEndpointRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetViewerTestEndpointResponse.SerializeToString,
        ),
        "GetDataScientistTestEndpoint": grpc.unary_unary_rpc_method_handler(
            servicer.GetDataScientistTestEndpoint,
            request_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDataScientistTestEndpointRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDataScientistTestEndpointResponse.SerializeToString,
        ),
        "GetDeveloperTestEndpoint": grpc.unary_unary_rpc_method_handler(
            servicer.GetDeveloperTestEndpoint,
            request_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDeveloperTestEndpointRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDeveloperTestEndpointResponse.SerializeToString,
        ),
        "GetAdminTestEndpoint": grpc.unary_unary_rpc_method_handler(
            servicer.GetAdminTestEndpoint,
            request_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAdminTestEndpointRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAdminTestEndpointResponse.SerializeToString,
        ),
        "GetOwnerTestEndpoint": grpc.unary_unary_rpc_method_handler(
            servicer.GetOwnerTestEndpoint,
            request_deserializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetOwnerTestEndpointRequest.FromString,
            response_serializer=chalk_dot_server_dot_v1_dot_authtesting__pb2.GetOwnerTestEndpointResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("chalk.server.v1.AuthTestingService", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class AuthTestingService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetUnauthedTestEndpoint(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.AuthTestingService/GetUnauthedTestEndpoint",
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetUnauthedTestEndpointRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetUnauthedTestEndpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAuthedTestEndpoint(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.AuthTestingService/GetAuthedTestEndpoint",
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAuthedTestEndpointRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAuthedTestEndpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetViewerTestEndpoint(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.AuthTestingService/GetViewerTestEndpoint",
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetViewerTestEndpointRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetViewerTestEndpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetDataScientistTestEndpoint(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.AuthTestingService/GetDataScientistTestEndpoint",
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDataScientistTestEndpointRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDataScientistTestEndpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetDeveloperTestEndpoint(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.AuthTestingService/GetDeveloperTestEndpoint",
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDeveloperTestEndpointRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetDeveloperTestEndpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetAdminTestEndpoint(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.AuthTestingService/GetAdminTestEndpoint",
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAdminTestEndpointRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetAdminTestEndpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetOwnerTestEndpoint(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/chalk.server.v1.AuthTestingService/GetOwnerTestEndpoint",
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetOwnerTestEndpointRequest.SerializeToString,
            chalk_dot_server_dot_v1_dot_authtesting__pb2.GetOwnerTestEndpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.prompt_pb2 import (
    CreateNamedPromptRequest,
    CreateNamedPromptResponse,
    CreatePromptEvaluationRunRequest,
    CreatePromptEvaluationRunResponse,
    GetNamedPromptRequest,
    GetNamedPromptResponse,
    GetPromptEvaluationRunRequest,
    GetPromptEvaluationRunResponse,
    ListNamedPromptsRequest,
    ListNamedPromptsResponse,
    ListPromptEvaluationRunsRequest,
    ListPromptEvaluationRunsResponse,
    ListPromptVariantsRequest,
    ListPromptVariantsResponse,
    UpdateNamedPromptRequest,
    UpdateNamedPromptResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class PromptServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    ListNamedPrompts: UnaryUnaryMultiCallable[
        ListNamedPromptsRequest,
        ListNamedPromptsResponse,
    ]
    """Return a list of named prompts"""
    GetNamedPrompt: UnaryUnaryMultiCallable[
        GetNamedPromptRequest,
        GetNamedPromptResponse,
    ]
    """Return a single named prompt with the latest prompt variant"""
    CreateNamedPrompt: UnaryUnaryMultiCallable[
        CreateNamedPromptRequest,
        CreateNamedPromptResponse,
    ]
    """Create a new named prompt and its corresponding prompt variant"""
    UpdateNamedPrompt: UnaryUnaryMultiCallable[
        UpdateNamedPromptRequest,
        UpdateNamedPromptResponse,
    ]
    """Modify a named prompt; may create a new prompt variant if variant settings changed"""
    ListPromptVariants: UnaryUnaryMultiCallable[
        ListPromptVariantsRequest,
        ListPromptVariantsResponse,
    ]
    """Return a list of prompt variants"""
    ListPromptEvaluationRuns: UnaryUnaryMultiCallable[
        ListPromptEvaluationRunsRequest,
        ListPromptEvaluationRunsResponse,
    ]
    """Return a list of prompt evaluation runs"""
    GetPromptEvaluationRun: UnaryUnaryMultiCallable[
        GetPromptEvaluationRunRequest,
        GetPromptEvaluationRunResponse,
    ]
    """Return a prompt evaluation run and its prompt variants"""
    CreatePromptEvaluationRun: UnaryUnaryMultiCallable[
        CreatePromptEvaluationRunRequest,
        CreatePromptEvaluationRunResponse,
    ]
    """Start a new prompt evaluation run"""

class PromptServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def ListNamedPrompts(
        self,
        request: ListNamedPromptsRequest,
        context: ServicerContext,
    ) -> ListNamedPromptsResponse:
        """Return a list of named prompts"""
    @abstractmethod
    def GetNamedPrompt(
        self,
        request: GetNamedPromptRequest,
        context: ServicerContext,
    ) -> GetNamedPromptResponse:
        """Return a single named prompt with the latest prompt variant"""
    @abstractmethod
    def CreateNamedPrompt(
        self,
        request: CreateNamedPromptRequest,
        context: ServicerContext,
    ) -> CreateNamedPromptResponse:
        """Create a new named prompt and its corresponding prompt variant"""
    @abstractmethod
    def UpdateNamedPrompt(
        self,
        request: UpdateNamedPromptRequest,
        context: ServicerContext,
    ) -> UpdateNamedPromptResponse:
        """Modify a named prompt; may create a new prompt variant if variant settings changed"""
    @abstractmethod
    def ListPromptVariants(
        self,
        request: ListPromptVariantsRequest,
        context: ServicerContext,
    ) -> ListPromptVariantsResponse:
        """Return a list of prompt variants"""
    @abstractmethod
    def ListPromptEvaluationRuns(
        self,
        request: ListPromptEvaluationRunsRequest,
        context: ServicerContext,
    ) -> ListPromptEvaluationRunsResponse:
        """Return a list of prompt evaluation runs"""
    @abstractmethod
    def GetPromptEvaluationRun(
        self,
        request: GetPromptEvaluationRunRequest,
        context: ServicerContext,
    ) -> GetPromptEvaluationRunResponse:
        """Return a prompt evaluation run and its prompt variants"""
    @abstractmethod
    def CreatePromptEvaluationRun(
        self,
        request: CreatePromptEvaluationRunRequest,
        context: ServicerContext,
    ) -> CreatePromptEvaluationRunResponse:
        """Start a new prompt evaluation run"""

def add_PromptServiceServicer_to_server(servicer: PromptServiceServicer, server: Server) -> None: ...



from __future__ import annotations

from typing_extensions import final

from chalk import StaticOperator


@final
class ParquetScanOperator(StaticOperator):
    _chalk__operator_name = StaticOperator._chalk__operator_prefix + "parquet_scan"

    def __init__(self, files: tuple[str, ...], column_names: tuple[str, ...], *, aws_role_arn: str | None = None):
        super().__init__(files=files, column_names=column_names, parent=None)
        self.column_names = column_names
        self.files = files
        self.aws_role_arn = aws_role_arn



from chalk._gen.chalk.utils.v1 import encoding_pb2 as _encoding_pb2
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class Permission(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PERMISSION_UNSPECIFIED: _ClassVar[Permission]
    PERMISSION_INSECURE_UNAUTHENTICATED: _ClassVar[Permission]
    PERMISSION_AUTHENTICATED: _ClassVar[Permission]
    PERMISSION_QUERY_ONLINE: _ClassVar[Permission]
    PERMISSION_QUERY_OFFLINE: _ClassVar[Permission]
    PERMISSION_MONITORING_CREATE: _ClassVar[Permission]
    PERMISSION_MONITORING_READ: _ClassVar[Permission]
    PERMISSION_TEAM_ADD: _ClassVar[Permission]
    PERMISSION_TEAM_DELETE: _ClassVar[Permission]
    PERMISSION_TEAM_LIST: _ClassVar[Permission]
    PERMISSION_TEAM_ADMIN: _ClassVar[Permission]
    PERMISSION_DEPLOY_READ: _ClassVar[Permission]
    PERMISSION_DEPLOY_CREATE: _ClassVar[Permission]
    PERMISSION_DEPLOY_PREVIEW: _ClassVar[Permission]
    PERMISSION_DEPLOY_REDEPLOY: _ClassVar[Permission]
    PERMISSION_LOGS_LIST: _ClassVar[Permission]
    PERMISSION_CRON_READ: _ClassVar[Permission]
    PERMISSION_CRON_CREATE: _ClassVar[Permission]
    PERMISSION_SECRETS_WRITE: _ClassVar[Permission]
    PERMISSION_SECRETS_DECRYPT: _ClassVar[Permission]
    PERMISSION_SECRETS_LIST: _ClassVar[Permission]
    PERMISSION_TOKENS_WRITE: _ClassVar[Permission]
    PERMISSION_TOKENS_LIST: _ClassVar[Permission]
    PERMISSION_MIGRATE_READ: _ClassVar[Permission]
    PERMISSION_MIGRATE_PLAN: _ClassVar[Permission]
    PERMISSION_MIGRATE_EXECUTE: _ClassVar[Permission]
    PERMISSION_PROJECT_CREATE: _ClassVar[Permission]
    PERMISSION_CHALK_ADMIN: _ClassVar[Permission]
    PERMISSION_BILLING_READ: _ClassVar[Permission]

PERMISSION_UNSPECIFIED: Permission
PERMISSION_INSECURE_UNAUTHENTICATED: Permission
PERMISSION_AUTHENTICATED: Permission
PERMISSION_QUERY_ONLINE: Permission
PERMISSION_QUERY_OFFLINE: Permission
PERMISSION_MONITORING_CREATE: Permission
PERMISSION_MONITORING_READ: Permission
PERMISSION_TEAM_ADD: Permission
PERMISSION_TEAM_DELETE: Permission
PERMISSION_TEAM_LIST: Permission
PERMISSION_TEAM_ADMIN: Permission
PERMISSION_DEPLOY_READ: Permission
PERMISSION_DEPLOY_CREATE: Permission
PERMISSION_DEPLOY_PREVIEW: Permission
PERMISSION_DEPLOY_REDEPLOY: Permission
PERMISSION_LOGS_LIST: Permission
PERMISSION_CRON_READ: Permission
PERMISSION_CRON_CREATE: Permission
PERMISSION_SECRETS_WRITE: Permission
PERMISSION_SECRETS_DECRYPT: Permission
PERMISSION_SECRETS_LIST: Permission
PERMISSION_TOKENS_WRITE: Permission
PERMISSION_TOKENS_LIST: Permission
PERMISSION_MIGRATE_READ: Permission
PERMISSION_MIGRATE_PLAN: Permission
PERMISSION_MIGRATE_EXECUTE: Permission
PERMISSION_PROJECT_CREATE: Permission
PERMISSION_CHALK_ADMIN: Permission
PERMISSION_BILLING_READ: Permission
PERMISSION_FIELD_NUMBER: _ClassVar[int]
permission: _descriptor.FieldDescriptor



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.scheduler_pb2 import (
    ManualTriggerScheduledQueryRequest,
    ManualTriggerScheduledQueryResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class SchedulerServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    ManualTriggerScheduledQuery: UnaryUnaryMultiCallable[
        ManualTriggerScheduledQueryRequest,
        ManualTriggerScheduledQueryResponse,
    ]

class SchedulerServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def ManualTriggerScheduledQuery(
        self,
        request: ManualTriggerScheduledQueryRequest,
        context: ServicerContext,
    ) -> ManualTriggerScheduledQueryResponse: ...

def add_SchedulerServiceServicer_to_server(servicer: SchedulerServiceServicer, server: Server) -> None: ...



from chalk._gen.chalk.arrow.v1 import arrow_pb2 as _arrow_pb2
from chalk._gen.chalk.artifacts.v1 import export_pb2 as _export_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.common.v1 import chalk_error_pb2 as _chalk_error_pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as _graph_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureSQL(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "deployment_id",
        "fqn",
        "name",
        "namespace",
        "max_staleness",
        "etl_offline_to_online",
        "description",
        "owner",
        "tags",
        "kind_enum",
        "kind",
        "was_reset",
        "internal_version",
        "is_singleton",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    FQN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    MAX_STALENESS_FIELD_NUMBER: _ClassVar[int]
    ETL_OFFLINE_TO_ONLINE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    KIND_ENUM_FIELD_NUMBER: _ClassVar[int]
    KIND_FIELD_NUMBER: _ClassVar[int]
    WAS_RESET_FIELD_NUMBER: _ClassVar[int]
    INTERNAL_VERSION_FIELD_NUMBER: _ClassVar[int]
    IS_SINGLETON_FIELD_NUMBER: _ClassVar[int]
    id: int
    environment_id: str
    deployment_id: str
    fqn: str
    name: str
    namespace: str
    max_staleness: str
    etl_offline_to_online: bool
    description: str
    owner: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    kind_enum: str
    kind: str
    was_reset: bool
    internal_version: int
    is_singleton: bool
    def __init__(
        self,
        id: _Optional[int] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
        fqn: _Optional[str] = ...,
        name: _Optional[str] = ...,
        namespace: _Optional[str] = ...,
        max_staleness: _Optional[str] = ...,
        etl_offline_to_online: bool = ...,
        description: _Optional[str] = ...,
        owner: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        kind_enum: _Optional[str] = ...,
        kind: _Optional[str] = ...,
        was_reset: bool = ...,
        internal_version: _Optional[int] = ...,
        is_singleton: bool = ...,
    ) -> None: ...

class GetFeatureSQLResponse(_message.Message):
    __slots__ = ("features",)
    FEATURES_FIELD_NUMBER: _ClassVar[int]
    features: _containers.RepeatedCompositeFieldContainer[FeatureSQL]
    def __init__(self, features: _Optional[_Iterable[_Union[FeatureSQL, _Mapping]]] = ...) -> None: ...

class GetFeatureSQLRequest(_message.Message):
    __slots__ = ("deployment_id",)
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    deployment_id: str
    def __init__(self, deployment_id: _Optional[str] = ...) -> None: ...

class FeatureMetadata(_message.Message):
    __slots__ = (
        "fqn",
        "name",
        "namespace",
        "description",
        "owner",
        "tags",
        "max_staleness",
        "etl_offline_to_online",
        "pa_dtype",
        "nullable",
    )
    FQN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    OWNER_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    MAX_STALENESS_FIELD_NUMBER: _ClassVar[int]
    ETL_OFFLINE_TO_ONLINE_FIELD_NUMBER: _ClassVar[int]
    PA_DTYPE_FIELD_NUMBER: _ClassVar[int]
    NULLABLE_FIELD_NUMBER: _ClassVar[int]
    fqn: str
    name: str
    namespace: str
    description: str
    owner: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    max_staleness: str
    etl_offline_to_online: bool
    pa_dtype: _arrow_pb2.ArrowType
    nullable: bool
    def __init__(
        self,
        fqn: _Optional[str] = ...,
        name: _Optional[str] = ...,
        namespace: _Optional[str] = ...,
        description: _Optional[str] = ...,
        owner: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        max_staleness: _Optional[str] = ...,
        etl_offline_to_online: bool = ...,
        pa_dtype: _Optional[_Union[_arrow_pb2.ArrowType, _Mapping]] = ...,
        nullable: bool = ...,
    ) -> None: ...

class GetFeaturesMetadataResponse(_message.Message):
    __slots__ = ("features", "environment_id", "deployment_id")
    FEATURES_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    features: _containers.RepeatedCompositeFieldContainer[FeatureMetadata]
    environment_id: str
    deployment_id: str
    def __init__(
        self,
        features: _Optional[_Iterable[_Union[FeatureMetadata, _Mapping]]] = ...,
        environment_id: _Optional[str] = ...,
        deployment_id: _Optional[str] = ...,
    ) -> None: ...

class GetFeaturesMetadataRequest(_message.Message):
    __slots__ = ("fqns_filter",)
    FQNS_FILTER_FIELD_NUMBER: _ClassVar[int]
    fqns_filter: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, fqns_filter: _Optional[_Iterable[str]] = ...) -> None: ...

class UpdateGraphRequest(_message.Message):
    __slots__ = ("deployment_id", "graph", "chalkpy_version", "tag", "export")
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    GRAPH_FIELD_NUMBER: _ClassVar[int]
    CHALKPY_VERSION_FIELD_NUMBER: _ClassVar[int]
    TAG_FIELD_NUMBER: _ClassVar[int]
    EXPORT_FIELD_NUMBER: _ClassVar[int]
    deployment_id: str
    graph: _graph_pb2.Graph
    chalkpy_version: str
    tag: str
    export: _export_pb2.Export
    def __init__(
        self,
        deployment_id: _Optional[str] = ...,
        graph: _Optional[_Union[_graph_pb2.Graph, _Mapping]] = ...,
        chalkpy_version: _Optional[str] = ...,
        tag: _Optional[str] = ...,
        export: _Optional[_Union[_export_pb2.Export, _Mapping]] = ...,
    ) -> None: ...

class UpdateGraphResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetGraphRequest(_message.Message):
    __slots__ = ("deployment_id",)
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    deployment_id: str
    def __init__(self, deployment_id: _Optional[str] = ...) -> None: ...

class GetGraphResponse(_message.Message):
    __slots__ = ("graph", "chalkpy_version", "tag", "export", "deployment_id")
    GRAPH_FIELD_NUMBER: _ClassVar[int]
    CHALKPY_VERSION_FIELD_NUMBER: _ClassVar[int]
    TAG_FIELD_NUMBER: _ClassVar[int]
    EXPORT_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    graph: _graph_pb2.Graph
    chalkpy_version: str
    tag: str
    export: _export_pb2.Export
    deployment_id: str
    def __init__(
        self,
        graph: _Optional[_Union[_graph_pb2.Graph, _Mapping]] = ...,
        chalkpy_version: _Optional[str] = ...,
        tag: _Optional[str] = ...,
        export: _Optional[_Union[_export_pb2.Export, _Mapping]] = ...,
        deployment_id: _Optional[str] = ...,
    ) -> None: ...

class PythonVersion(_message.Message):
    __slots__ = ("major", "minor", "patch")
    MAJOR_FIELD_NUMBER: _ClassVar[int]
    MINOR_FIELD_NUMBER: _ClassVar[int]
    PATCH_FIELD_NUMBER: _ClassVar[int]
    major: int
    minor: int
    patch: int
    def __init__(
        self, major: _Optional[int] = ..., minor: _Optional[int] = ..., patch: _Optional[int] = ...
    ) -> None: ...

class GetCodegenFeaturesFromGraphRequest(_message.Message):
    __slots__ = ("deployment_id", "branch", "python_version")
    DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    BRANCH_FIELD_NUMBER: _ClassVar[int]
    PYTHON_VERSION_FIELD_NUMBER: _ClassVar[int]
    deployment_id: str
    branch: str
    python_version: PythonVersion
    def __init__(
        self,
        deployment_id: _Optional[str] = ...,
        branch: _Optional[str] = ...,
        python_version: _Optional[_Union[PythonVersion, _Mapping]] = ...,
    ) -> None: ...

class GetCodegenFeaturesFromGraphResponse(_message.Message):
    __slots__ = ("codegen", "errors")
    CODEGEN_FIELD_NUMBER: _ClassVar[int]
    ERRORS_FIELD_NUMBER: _ClassVar[int]
    codegen: str
    errors: _containers.RepeatedCompositeFieldContainer[_chalk_error_pb2.ChalkError]
    def __init__(
        self,
        codegen: _Optional[str] = ...,
        errors: _Optional[_Iterable[_Union[_chalk_error_pb2.ChalkError, _Mapping]]] = ...,
    ) -> None: ...



from __future__ import annotations

import base64
import collections.abc
import dataclasses
import decimal
import enum
import ipaddress
import uuid
from datetime import date, datetime, time, timedelta
from typing import Any, FrozenSet, List, Set, Tuple, Type, TypeVar, Union, cast

import attrs
import cattrs
import dateutil.parser
import isodate

# Need the *real* BaseModel here (not the v1 one) so that we can register the right hook
from pydantic import BaseModel as BaseBaseModel

try:
    from pydantic.v1 import BaseModel as V1BaseModel
except ImportError:
    V1BaseModel = None

from typing_extensions import get_args, get_origin, is_typeddict

from chalk.features._encoding.primitive import ChalkStructType, TPrimitive
from chalk.utils.cached_type_hints import cached_get_type_hints
from chalk.utils.collections import is_namedtuple, unwrap_optional_and_annotated_if_needed
from chalk.utils.enum import get_enum_value_type
from chalk.utils.json import JSON
from chalk.utils.pydanticutil.pydantic_compat import construct_pydantic_model, is_pydantic_basemodel

TRich = TypeVar("TRich")
T = TypeVar("T")

# Converter to go from TPython to a TPrimitive
_rich_converter = cattrs.Converter()

#########
# Structs
#########

_rich_converter.register_unstructure_hook_func(
    is_namedtuple, lambda x: {k: _rich_converter.unstructure(v) for (k, v) in x._asdict().items()}
)
_rich_converter.register_unstructure_hook_func(
    dataclasses.is_dataclass,
    lambda x: {field.name: _rich_converter.unstructure(getattr(x, field.name)) for field in dataclasses.fields(x)},
)
_rich_converter.register_unstructure_hook(
    BaseBaseModel, lambda x: {k: _rich_converter.unstructure(v) for (k, v) in x.dict().items()}
)
if V1BaseModel is not None:
    _rich_converter.register_unstructure_hook(
        V1BaseModel, lambda x: {k: _rich_converter.unstructure(v) for (k, v) in x.dict().items()}
    )
_rich_converter.register_unstructure_hook_func(
    attrs.has,
    lambda x: {k: _rich_converter.unstructure(v) for (k, v) in attrs.asdict(x, recurse=False).items()},
)


def _is_struct(typ: Type):
    typ = unwrap_optional_and_annotated_if_needed(typ)
    return (
        is_namedtuple(typ)
        or dataclasses.is_dataclass(typ)
        or is_typeddict(typ)
        or (isinstance(typ, type) and is_pydantic_basemodel(typ))
        or attrs.has(typ)
    )


def _structure_struct(obj: Any, typ: Type):
    if not isinstance(obj, (collections.abc.Sequence, collections.abc.Mapping)):
        # If an already-constructed class is passed in, then we still need to restructure it
        # to ensure that any recursive fields are properly structured
        obj = _rich_converter.unstructure(obj)
    typ = unwrap_optional_and_annotated_if_needed(typ)
    type_hints = cached_get_type_hints(typ)
    if obj is None:
        obj = {k: None for k in type_hints}
    if isinstance(obj, collections.abc.Sequence):
        obj = {k: v for (k, v) in zip(type_hints.keys(), obj)}
    if not isinstance(obj, collections.abc.Mapping):
        raise TypeError(f"Unable to structure object `{obj}` into type `{typ}`. Only dictionaries are supported.")
    kwargs = {k: _rich_converter.structure(v, type_hints[k]) for (k, v) in obj.items()}
    if isinstance(typ, type) and is_pydantic_basemodel(typ):  # pyright: ignore[reportUnnecessaryIsInstance]
        # Using .construct to bypass pydantic validation, which could fail if there are null values in kwargs
        # and the field is not annotated as null
        return construct_pydantic_model(typ, **kwargs)
    else:
        return typ(**kwargs)


def is_chalk_struct(typ: Type):
    return isinstance(typ, ChalkStructType)


def _structure_chalk_struct(obj: Any, typ: Type):
    if not isinstance(typ, ChalkStructType):
        raise TypeError(f"Expected ChalkStructType instead of {typ}")
    type_hints = typ.__chalk_type_hints__
    if obj is None:
        obj = {k: None for k in type_hints}
    if isinstance(obj, collections.abc.Sequence):
        obj = {k: v for (k, v) in zip(type_hints.keys(), obj)}
    if not isinstance(obj, collections.abc.Mapping):
        raise TypeError(f"Unable to structure object `{obj}` into type `{typ}`. Only dictionaries are supported.")
    return {k: _rich_converter.structure(v, type_hints[k]) for (k, v) in obj.items()}


_rich_converter.register_structure_hook_func(_is_struct, _structure_struct)
_rich_converter.register_structure_hook_func(is_chalk_struct, _structure_chalk_struct)

#############
# Collections
#############

# Unlike lists and tuples which are ordered, sets are unordered
# Thus, when converting to a list, the order must be consistent
# To do so, sort all the values
_rich_converter.register_unstructure_hook(set, lambda x: sorted(list(x)))
_rich_converter.register_unstructure_hook(frozenset, lambda x: sorted(list(x)))


# Lists and tuples do not need an unstructure hook -- the default is fine


def _structure_collection(
    obj: Any, typ: Union[Type[FrozenSet[T]], Type[Set[T]], Type[List[T]], Type[Tuple[T, ...]]]
) -> Union[FrozenSet[T], Set[T], List[T], Tuple[T, ...]]:
    origin = get_origin(typ)
    if origin in (set, Set):
        constructor = set
    elif origin in (frozenset, FrozenSet):
        constructor = frozenset
    elif origin in (list, List):
        constructor = list
    elif origin in (tuple, Tuple):
        constructor = tuple
    else:
        raise TypeError(f"Unsupported set type: {typ}")
    args = get_args(typ)
    if len(args) < 1:
        raise TypeError(
            f"{typ} types must be parameterized with the type of the contained value -- for example, `{typ}[int]`"
        )
    if len(args) > 1:
        if origin in (tuple, Tuple):
            if len(args) != 2 and args[1] != ...:
                raise TypeError(
                    (
                        "Tuple features must have a fixed type and be variable-length tuples (e.g. `Tuple[int, ...]`). "
                        " If you would like a fixed-length of potentially different types, used a NamedTuple."
                    )
                )
        else:
            raise TypeError(f"{typ} should be parameterized with only one type")

    inner_typ = args[0]
    if obj is None:
        return cast(Union[FrozenSet[T], Set[T], List[T], Tuple[T, ...]], None)
    if not isinstance(obj, (collections.abc.Set, collections.abc.Sequence)):
        raise TypeError(f"Cannot structure '{obj}' into a `{typ}`")
    return constructor(_rich_converter.structure(x, inner_typ) for x in obj)


def _is_collection(typ: Type):
    origin = get_origin(typ)
    return origin in (
        set,
        Set,
        frozenset,
        FrozenSet,
        list,
        List,
        tuple,
        Tuple,
    )


_rich_converter.register_structure_hook_func(_is_collection, _structure_collection)

#######
# Enums
#######

_rich_converter.register_unstructure_hook(enum.Enum, lambda x: _rich_converter.unstructure(x.value))


def _structure_enum(obj: Any, typ: Type[enum.Enum]) -> enum.Enum:
    if isinstance(obj, typ):
        return obj
    if obj is None:
        return cast(enum.Enum, None)
    enum_typ = get_enum_value_type(typ)
    try:
        return typ(_rich_converter.structure(obj, enum_typ))
    except (TypeError, ValueError):
        pass
    if isinstance(obj, str):
        try:
            return typ[obj]
        except KeyError:
            pass
    allowed_values = ", ".join(f"'{x}'" for x in typ.__members__.values())
    raise ValueError(f"Cannot convert '{obj}' to Enum `{typ}`. Possible values are: {allowed_values}")


_rich_converter.register_structure_hook(enum.Enum, _structure_enum)

##########
# Decimals
##########

_rich_converter.register_unstructure_hook(decimal.Decimal, lambda x: str(x.normalize()))


def _structure_decimal(obj: Any, typ: Type[decimal.Decimal]) -> decimal.Decimal:
    if obj is None:
        return cast(decimal.Decimal, None)
    if isinstance(obj, decimal.Decimal):
        return obj
    return decimal.Decimal(obj)


_rich_converter.register_structure_hook(decimal.Decimal, _structure_decimal)

#######
# Bytes
#######

_rich_converter.register_unstructure_hook(bytes, lambda x: x)


def _structure_bytes(x: Any, typ: Type[bytes]) -> bytes:
    if x is None:
        return cast(bytes, None)
    if isinstance(x, bytes):
        return x
    if isinstance(x, str):
        return base64.b64decode(x)
    raise TypeError(f"Cannot structure {x} into bytes")


_rich_converter.register_structure_hook(bytes, _structure_bytes)

##########
# Duration
##########

_rich_converter.register_unstructure_hook(timedelta, lambda x: x)


def _structure_timedelta(x: Any, typ: Type[timedelta]) -> timedelta:
    if x is None:
        return cast(timedelta, None)
    if isinstance(x, timedelta):
        return x
    if isinstance(x, str):
        return isodate.parse_duration(x)
    raise TypeError(f"Cannot structure {x} into a duration")


_rich_converter.register_structure_hook(timedelta, _structure_timedelta)

######
# Date
######

_rich_converter.register_unstructure_hook_func(
    lambda x: isinstance(x, date) and not isinstance(x, datetime), lambda x: x
)


def _structure_date(x: Any, typ: Type[date]) -> date:
    if x is None:
        return cast(date, None)
    if isinstance(x, datetime):
        if x.time() != time():
            raise TypeError(f"Datetime '{x}' has a non-zero time component, which cannot be safely cast into a date")
        return x.date()
    if isinstance(x, date):
        return x
    if isinstance(x, str):
        return isodate.parse_date(x)
    raise TypeError(f"Cannot structure {x} into a date")


_rich_converter.register_structure_hook_func(
    lambda x: isinstance(x, type)  # pyright: ignore[reportUnnecessaryIsInstance]
    and issubclass(x, date)  # pyright: ignore[reportUnnecessaryIsInstance]
    and not issubclass(x, datetime),
    _structure_date,
)

######
# Time
######

_rich_converter.register_unstructure_hook(time, lambda x: x)


def _structure_time(x: Any, typ: Type[time]) -> time:
    if x is None:
        return cast(time, None)
    if isinstance(x, time):
        return x
    if isinstance(x, str):
        return isodate.parse_time(x)
    raise TypeError(f"Cannot structure {x} into a time")


_rich_converter.register_structure_hook(time, _structure_time)

##########
# Datetime
##########

_rich_converter.register_unstructure_hook(datetime, lambda x: x)


def _structure_datetime(x: Any, typ: Type[datetime]) -> datetime:
    if x is None:
        return cast(datetime, None)
    if isinstance(x, str):
        x = dateutil.parser.parse(x)
    if isinstance(x, datetime):
        return x
    if isinstance(x, date):
        return datetime.combine(x, time())
    raise TypeError(f"Cannot structure {x} into a datetime")


_rich_converter.register_structure_hook(datetime, _structure_datetime)

##########
# UUID
##########

_rich_converter.register_unstructure_hook(uuid.UUID, lambda x: x)


def _structure_uuid(x: Any, typ: Type[uuid.UUID]) -> uuid.UUID:
    if x is None:
        return cast(uuid.UUID, None)
    if isinstance(x, uuid.UUID):
        return x
    if isinstance(x, str):
        return uuid.UUID(x)
    raise TypeError(f"Cannot structure {x} into a uuid.UUID")


_rich_converter.register_structure_hook(uuid.UUID, _structure_uuid)

"""
ipaddress.IPv4Network
"""
_rich_converter.register_unstructure_hook(ipaddress.IPv4Address, lambda x: x)


def _structure_ipv4(x: Any, typ: Type[ipaddress.IPv4Address]) -> ipaddress.IPv4Address:
    if x is None:
        return cast(ipaddress.IPv4Address, None)
    if isinstance(x, ipaddress.IPv4Address):
        return x
    if isinstance(x, str):
        return ipaddress.IPv4Address(x)
    if isinstance(x, int):
        return ipaddress.IPv4Address(x)
    raise TypeError(f"Cannot structure {x} into an ipaddress.IPv4Address")


_rich_converter.register_structure_hook(ipaddress.IPv4Address, _structure_ipv4)

"""
ipaddress.IPv6Network
"""
_rich_converter.register_unstructure_hook(ipaddress.IPv6Address, lambda x: x)


def _structure_ipv6(x: Any, typ: Type[ipaddress.IPv6Address]) -> ipaddress.IPv6Address:
    if x is None:
        return cast(ipaddress.IPv6Address, None)
    if isinstance(x, ipaddress.IPv6Address):
        return x
    if isinstance(x, str):
        return ipaddress.IPv6Address(x)
    if isinstance(x, int):
        return ipaddress.IPv6Address(x)
    raise TypeError(f"Cannot structure {x} into an ipaddress.IPv6Address")


_rich_converter.register_structure_hook(ipaddress.IPv6Address, _structure_ipv6)

#############
# Numpy types
#############
try:
    import numpy as np
except ImportError:
    pass
else:
    _rich_converter.register_unstructure_hook_func(lambda x: np.issubdtype(x, np.bool_), bool)
    _rich_converter.register_unstructure_hook_func(lambda x: np.issubdtype(x, np.integer), int)
    _rich_converter.register_unstructure_hook_func(lambda x: np.issubdtype(x, np.floating), float)


##############
# Rich types
##############


def _structure_rich(obj: Any, typ: Type):
    if obj is None:
        # Always allow None, even if the field is non-optional
        return None
    if issubclass(typ, enum.Enum):
        return _structure_enum(obj, typ)
    if isinstance(obj, typ):
        return obj
    if issubclass(typ, bool):
        # For booleans, we don't want to use the bool(...) constructor, as
        # that doesn't handle strings like y/yes/n/no/t/true/f/false
        if isinstance(obj, str):
            if obj.lower() in ("y", "yes", "t", "true", "1"):
                return True
            if obj.lower() in ("n", "no", "f", "false", "0"):
                return False
        if isinstance(obj, int):
            if obj == 0:
                return False
            if obj == 1:
                return True
        if np and isinstance(obj, np.bool_):
            return bool(obj)
        raise ValueError(
            f"Cannot convert {obj} to a boolean. Allowed values are y, yes, t, true, 1, n, no, f, false, 0."
        )

    if (
        isinstance(obj, (list, set, frozenset, dict, tuple, BaseBaseModel))
        or (V1BaseModel is not None and isinstance(obj, V1BaseModel))
        or dataclasses.is_dataclass(obj)
        or attrs.has(type(obj))
    ):
        raise ValueError(f"Object {obj} of type {type(obj).__name__} cannot be converted into a {typ.__name__}")

    # Perform this import lazily, to prevent a circular import.
    from chalk.features.underscore import Underscore

    if isinstance(obj, Underscore) and not issubclass(typ, Underscore):
        raise ValueError(
            f"Chalk expression '{repr(obj)}' cannot be provided where a concrete value of type '{typ.__name__}' is expected"
        )

    return typ(obj)  # pyright: ignore


_rich_converter.register_structure_hook(int, _structure_rich)
_rich_converter.register_structure_hook(float, _structure_rich)
_rich_converter.register_structure_hook(bytes, _structure_rich)
_rich_converter.register_structure_hook(str, _structure_rich)
_rich_converter.register_structure_hook(bool, _structure_rich)


def unstructure_rich_to_primitive(val: Any) -> TPrimitive:
    if val is None or type(val) in (str, int, float, bytes, time, datetime, date, timedelta, bool):
        # Short-circuit these rich types, as the primitive types are the same
        return val
    # Special case for enums, particularly enums that use a mixin
    # i.e. class MyEnum(str, enum.Enum)
    # cattrs seems unable to detect that the class inherits from Enum so doesn't call the correct serde hooks
    if isinstance(val, enum.Enum):
        return _rich_converter.unstructure(val.value)
    return _rich_converter.unstructure(val)


def structure_primitive_to_rich(val: TPrimitive, typ: Type[TRich]) -> TRich:
    if (
        (val is None or type(val) is typ)
        and typ
        in (
            str,
            int,
            bool,
            float,
            bytes,
            timedelta,
            decimal.Decimal,
            date,
            datetime,
            time,
        )
    ) or typ is JSON:  # pyright: ignore[reportUnnecessaryComparison]
        # Short circuit these rich types
        return cast(TRich, val)
    return _rich_converter.structure(val, typ)
